from eltetrado.cli import eltetrado_cli, has_tetrad_cli


def test_has_tetrad_with_dssr(capfd):
    has_tetrad_cli(
        [
            "--input",
            "tests/files/1v3p-assembly-1.cif.gz",
            "--external-files",
            "tests/files/1v3p-assembly-1.json",
            "--tool",
            "dssr",
        ]
    )
    out, err = capfd.readouterr()
    assert out == "True\n"


def test_has_tetrad_with_fr3d(capfd):
    has_tetrad_cli(
        [
            "--input",
            "tests/files/1v3p-assembly-1.cif.gz",
            "--external-files",
            "tests/files/1v3p-assembly-1.txt",
            "--tool",
            "fr3d",
        ]
    )
    out, err = capfd.readouterr()
    assert out == "True\n"


def test_has_tetrad(capfd):
    has_tetrad_cli(
        [
            "--input",
            "tests/files/1v3p-assembly-1.cif.gz",
        ]
    )
    out, err = capfd.readouterr()
    assert out == "True\n"


def test_eltetrado_with_dssr(capfd):
    eltetrado_cli(
        [
            "--input",
            "tests/files/2awe-assembly-1.cif.gz",
            "--external-files",
            "tests/files/2awe-assembly-1.json",
            "--tool",
            "dssr",
            "--no-image",
        ]
    )
    out, err = capfd.readouterr()
    with open("tests/files/2awe-assembly-1.out") as f:
        assert out == f.read()


def test_eltetrado_with_fr3d(capfd):
    eltetrado_cli(
        [
            "--input",
            "tests/files/2awe-assembly-1.cif.gz",
            "--external-files",
            "tests/files/2awe-assembly-1.txt",
            "--tool",
            "fr3d",
            "--no-image",
        ]
    )
    out, err = capfd.readouterr()
    with open("tests/files/2awe-assembly-1-fr3d.out") as f:
        assert out == f.read()


def test_eltetrado_without_dssr(capfd):
    eltetrado_cli(
        [
            "--input",
            "tests/files/6fc9-assembly-1.cif.gz",
            "--no-image",
        ]
    )
    out, err = capfd.readouterr()
    with open("tests/files/6fc9-assembly-1.out") as f:
        assert out == f.read()



from pathlib import Path

import rnapolis.parser
from rnapolis.adapter import ExternalTool, parse_external_output

from eltetrado.analysis import eltetrado
from eltetrado.cli import handle_input_file


def test_5zev_tracts():
    """
    Analyse 5ZEV structure with DSSR-generated JSON and verify the detected
    G-quadruplex tracts.
    """
    cif_path = Path("tests/files/5zev-assembly1.cif.gz")
    json_path = Path("tests/files/5zev-assembly1.json")

    # Load the 3D structure using the same helper routine as the CLI
    cif_or_pdb = handle_input_file(str(cif_path))
    structure3d = rnapolis.parser.read_3d_structure(
        cif_or_pdb, 1, nucleic_acid_only=False
    )

    # Parse external interactions produced by DSSR
    base_interactions = parse_external_output(
        [str(json_path)], ExternalTool.DSSR, structure3d
    )

    # Run ElTetrado analysis
    analysis = eltetrado(
        base_interactions,
        structure3d,
        strict=False,
        no_reorder=False,
        stacking_mismatch=2,
    )

    # Expected tracts (in 5'→3' order within each tract)
    expected_tracts = {
        ("A.DG9", "A.DG10", "A.DG11"),
        ("A.DG21", "A.DG22", "A.DG23"),
        ("A.DG13", "A.DG15", "A.DG16"),
        ("A.DG2", "A.DG1", "A.DG20"),
    }

    # Collect tracts detected by ElTetrado
    detected_tracts = {
        tuple(nt.full_name for nt in tract.nucleotides)
        for helix in analysis.helices
        for quadruplex in helix.quadruplexes
        for tract in quadruplex.tracts
    }

    for tract in expected_tracts:
        assert tract in detected_tracts, (
            f"Expected tract {tract} not found; detected tracts: {detected_tracts}"
        )



import argparse
import gzip
import logging
import os
import sys
import tempfile
from typing import IO, List, Optional

import orjson
import rnapolis.annotator
import rnapolis.parser
from rnapolis.adapter import ExternalTool, parse_external_output, auto_detect_tool
from rnapolis.annotator import BaseInteractions, LeontisWesthof
from rnapolis.common import BasePair, Residue, Stacking
from rnapolis.tertiary import Structure3D

from eltetrado.analysis import Visualizer, eltetrado, has_tetrad
from eltetrado.dto import generate_dto


def eltetrado_cli(args=sys.argv[1:]):
    with open(os.path.join(os.path.dirname(__file__), "VERSION")) as f:
        version = f.read().strip()

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", help="path to input PDB or PDBx/mmCIF file")
    parser.add_argument("-o", "--output", help="(optional) path for output JSON file")
    parser.add_argument(
        "-m", "--model", help="(optional) model number to process", default=1, type=int
    )
    parser.add_argument(
        "--stacking-mismatch",
        help="a perfect tetrad stacking covers 4 nucleotides; this option can be used with value 1 or "
        "2 to allow this number of nucleotides to be non-stacked with otherwise well aligned "
        "tetrad [default=2]",
        default=2,
        type=int,
    )
    parser.add_argument(
        "--strict",
        action="store_true",
        help="nucleotides in tetrad are found when linked only by cWH pairing",
    )
    parser.add_argument(
        "--no-reorder",
        action="store_true",
        help="chains of bi- and tetramolecular quadruplexes should be reordered to be able to have "
        "them classified; when this is set, chains will be processed in original order, which for "
        "bi-/tetramolecular means that they will likely be misclassified; use with care!",
    )
    parser.add_argument(
        "--complete-2d",
        action="store_true",
        help="when set, the visualization will also show canonical base pairs to provide context for "
        "the quadruplex",
    )
    parser.add_argument(
        "--no-image",
        action="store_true",
        help="when set, the visualization will not be created at all",
    )
    parser.add_argument(
        "-e",
        "--external-files",
        nargs="*",
        default=[],
        help="path(s) to external tool output file(s); if omitted ElTetrado will compute interactions itself",
    )
    parser.add_argument(
        "--tool",
        choices=[t.value for t in ExternalTool],
        help="name of the external tool that produced the files (auto-detected when not provided)",
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s {}".format(version)
    )
    args = parser.parse_args(args)

    if not args.input:
        print(parser.print_help())
        sys.exit(1)

    cif_or_pdb = handle_input_file(args.input)
    structure3d = rnapolis.parser.read_3d_structure(
        cif_or_pdb, args.model, nucleic_acid_only=False
    )

    external_files: List[str] = args.external_files
    selected_tool: Optional[ExternalTool] = (
        ExternalTool(args.tool) if args.tool else None
    )

    if selected_tool is None and external_files:
        selected_tool = auto_detect_tool(external_files)
        logging.info(f"Auto-detected external tool: {selected_tool.value}")

    if selected_tool:
        if not external_files and selected_tool == ExternalTool.MAXIT:
            external_files = [args.input]
        base_interactions = parse_external_output(
            external_files, selected_tool, structure3d
        )
    else:
        base_interactions = rnapolis.annotator.extract_base_interactions(
            structure3d, args.model
        )

    analysis = eltetrado(
        base_interactions,
        structure3d,
        args.strict,
        args.no_reorder,
        args.stacking_mismatch,
    )
    print(analysis)

    if not args.no_image:
        visualizer = Visualizer(
            analysis, analysis.tetrads, args.complete_2d, analysis.global_index
        )

        basename = os.path.basename(args.input)
        root, ext = os.path.splitext(basename)
        if ext == ".gz":
            root, ext = os.path.splitext(root)
        prefix = root
        suffix = "str"
        visualizer.visualize(prefix, suffix)

        for i, helix in enumerate(analysis.helices):
            hv = Visualizer(
                analysis, helix.tetrads, args.complete_2d, analysis.global_index
            )
            suffix = "h{}".format(i + 1)
            hv.visualize(prefix, suffix)

            for j, quadruplex in enumerate(helix.quadruplexes):
                qv = Visualizer(
                    analysis,
                    quadruplex.tetrads,
                    args.complete_2d,
                    analysis.global_index,
                )
                qv.visualize(prefix, "{}-q{}".format(suffix, j + 1))

                for k, tetrad in enumerate(quadruplex.tetrads):
                    tv = Visualizer(
                        analysis, [tetrad], args.complete_2d, analysis.global_index
                    )
                    tv.visualize(prefix, "{}-q{}-t{}".format(suffix, j + 1, k + 1))

    if args.output:
        dto = generate_dto(analysis)

        with open(args.output, "wb") as jsonfile:
            jsonfile.write(orjson.dumps(dto))


def has_tetrad_cli(args=sys.argv[1:]):
    with open(os.path.join(os.path.dirname(__file__), "VERSION")) as f:
        version = f.read().strip()

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", help="path to input PDB or PDBx/mmCIF file")
    parser.add_argument(
        "-m", "--model", help="(optional) model number to process", default=1, type=int
    )
    parser.add_argument(
        "-e",
        "--external-files",
        nargs="*",
        default=[],
        help="path(s) to external tool output file(s); if omitted ElTetrado will compute interactions itself",
    )
    parser.add_argument(
        "--tool",
        choices=[t.value for t in ExternalTool],
        help="name of the external tool that produced the files (auto-detected when not provided)",
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s {}".format(version)
    )
    args = parser.parse_args(args)

    if not args.input:
        print(parser.print_help())
        sys.exit(1)

    cif_or_pdb = handle_input_file(args.input)
    structure3d = rnapolis.parser.read_3d_structure(
        cif_or_pdb, args.model, nucleic_acid_only=False
    )

    external_files: List[str] = args.external_files
    selected_tool: Optional[ExternalTool] = (
        ExternalTool(args.tool) if args.tool else None
    )

    if selected_tool is None and external_files:
        selected_tool = auto_detect_tool(external_files)
        logging.info(f"Auto-detected external tool: {selected_tool.value}")

    if selected_tool:
        if not external_files and selected_tool == ExternalTool.MAXIT:
            external_files = [args.input]
        base_interactions = parse_external_output(
            external_files, selected_tool, structure3d
        )
    else:
        base_interactions = rnapolis.annotator.extract_base_interactions(
            structure3d, args.model
        )

    print(has_tetrad(base_interactions, structure3d))


def handle_input_file(path) -> IO[str]:
    root, ext = os.path.splitext(path)

    if ext == ".gz":
        root, ext = os.path.splitext(root)
        file = tempfile.NamedTemporaryFile("w+", suffix=ext)
        with gzip.open(path, "rt") as f:
            file.write(f.read())
            file.seek(0)
    else:
        file = tempfile.NamedTemporaryFile("w+", suffix=ext)
        with open(path) as f:
            file.write(f.read())
            file.seek(0)
    return file


def read_secondary_structure_from_dssr(
    structure3d: Structure3D, model: int, dssr_json_path: str
) -> BaseInteractions:
    base_pairs: List[BasePair] = []
    stackings: List[Stacking] = []

    with open(dssr_json_path) as f:
        dssr = orjson.loads(f.read())

    for result in dssr.get("models", []):
        if result.get("model", None) == model:
            dssr = result.get("parameters", {})
            break

    for pair in dssr.get("pairs", []):
        nt1 = match_dssr_name_to_residue(structure3d, pair.get("nt1", None))
        nt2 = match_dssr_name_to_residue(structure3d, pair.get("nt2", None))
        lw = match_dssr_lw(pair.get("LW", None))

        if nt1 is not None and nt2 is not None and lw is not None:
            base_pairs.append(BasePair(nt1, nt2, lw, None))

    for stack in dssr.get("stacks", []):
        nts = [
            match_dssr_name_to_residue(structure3d, nt)
            for nt in stack.get("nts_long", "").split(",")
        ]
        for i in range(1, len(nts)):
            nt1 = nts[i - 1]
            nt2 = nts[i]
            if nt1 is not None and nt2 is not None:
                stackings.append(Stacking(nt1, nt2, None))

    return BaseInteractions(base_pairs, stackings, [], [], [])


def match_dssr_name_to_residue(
    structure3d: Structure3D, nt_id: Optional[str]
) -> Optional[Residue]:
    if nt_id is not None:
        nt_id = nt_id.split(":")[-1]
        for residue in structure3d.residues:
            if residue.full_name == nt_id:
                return residue
        logging.warn(f"Failed to find residue {nt_id}")
    return None


def match_dssr_lw(lw: Optional[str]) -> Optional[LeontisWesthof]:
    return LeontisWesthof[lw] if lw in dir(LeontisWesthof) else None


if __name__ == "__main__":
    eltetrado_cli()



import math
import re
from collections import Counter
from dataclasses import dataclass
from typing import List, Optional

from rnapolis.common import GlycosidicBond

from eltetrado.analysis import ONZ, Analysis, LoopType, Quadruplex, TetradPair
from eltetrado.model import Ion


@dataclass
class MetalDTO:
    symbol: str
    count: int


@dataclass
class NucleotideDTO:
    index: int
    chain: str
    number: int
    icode: Optional[str]
    molecule: str
    fullName: str
    shortName: str
    chi: float
    glycosidicBond: Optional[str]


@dataclass
class BasePairDTO:
    nt1: str
    nt2: str
    lw: str
    inTetrad: bool
    canonical: bool


@dataclass
class IonOutsideDTO:
    nt: str
    ion: str


@dataclass
class TetradDTO:
    id: str
    nt1: str
    nt2: str
    nt3: str
    nt4: str
    onz: str
    gbaClassification: Optional[str]
    planarityDeviation: float
    ionsChannel: List[str]
    ionsOutside: List[IonOutsideDTO]


@dataclass
class TetradPairDTO:
    tetrad1: str
    tetrad2: str
    direction: str
    rise: float
    twist: float


@dataclass
class LoopDTO:
    type: Optional[str]
    nucleotides: List[str]


@dataclass
class LoopClassificationDTO:
    classification: str
    loopProgression: str


@dataclass
class QuadruplexDTO:
    tetrads: List[TetradDTO]
    onzm: Optional[str]
    loopClassification: Optional[LoopClassificationDTO]
    gbaClassification: List[str]
    tracts: List[List[str]]
    bulges: List[str]
    loops: List[LoopDTO]


@dataclass
class HelixDTO:
    quadruplexes: List[QuadruplexDTO]
    tetradPairs: List[TetradPairDTO]


@dataclass
class TwoLineDotBracketDTO:
    sequence: str
    line1: str
    line2: str


@dataclass
class QuadruplexDotBracketDTO:
    sequence: str
    structure: str
    chi: str
    loop: str


@dataclass
class ResultDTO:
    metals: List[MetalDTO]
    nucleotides: List[NucleotideDTO]
    basePairs: List[BasePairDTO]
    helices: List[HelixDTO]
    dotBracket: TwoLineDotBracketDTO
    quadruplexDotBracket: QuadruplexDotBracketDTO


def convert_metals(analysis: Analysis) -> List[MetalDTO]:
    counter = Counter(map(lambda atom: atom.name.title(), analysis.ions))
    return [MetalDTO(Ion[k.title()].value, v) for k, v in counter.items()]


def convert_nucleotides(analysis: Analysis) -> List[NucleotideDTO]:
    return [
        NucleotideDTO(
            analysis.global_index[nt],
            nt.chain,
            nt.number,
            nt.icode,
            nt.molecule_type.value,
            nt.full_name,
            nt.one_letter_name,
            math.degrees(nt.chi),
            nt.chi_class.value if nt.chi_class else None,
        )
        for nt in analysis.structure3d.residues
        if nt.is_nucleotide
    ]


def convert_base_pairs(analysis: Analysis) -> List[BasePairDTO]:
    return [
        BasePairDTO(
            bp.nt1.full_name,
            bp.nt2.full_name,
            bp.lw.value,
            analysis.is_basepair_in_tetrad(bp),
            bp.is_canonical,
        )
        for bp in analysis.mapping.base_pairs
        if analysis.global_index[bp.nt1_3d] < analysis.global_index[bp.nt2_3d]
    ]


def convert_tetrads(quadruplex: Quadruplex) -> List[TetradDTO]:
    id_ = (
        lambda tetrad: f"{tetrad.nt1.full_name}-{tetrad.nt2.full_name}-{tetrad.nt3.full_name}-{tetrad.nt4.full_name}"
    )
    ions_channel = lambda tetrad: [
        Ion[atom.name.title()].value for atom in tetrad.ions_channel
    ]
    ions_outside = lambda tetrad: [
        IonOutsideDTO(nt.full_name, Ion[atom.name.title()].value)
        for nt, atoms in tetrad.ions_outside.items()
        for atom in atoms
    ]
    return [
        TetradDTO(
            id_(tetrad),
            tetrad.nt1.full_name,
            tetrad.nt2.full_name,
            tetrad.nt3.full_name,
            tetrad.nt4.full_name,
            tetrad.onz.value,
            tetrad.gba_class.value if tetrad.gba_class is not None else None,
            float(tetrad.planarity_deviation),
            ions_channel(tetrad),
            ions_outside(tetrad),
        )
        for tetrad in quadruplex.tetrads
    ]


def convert_tetrad_pairs(tetrad_pairs: List[TetradPair]) -> List[TetradPairDTO]:
    id_ = (
        lambda tetrad: f"{tetrad.nt1.full_name}-{tetrad.nt2.full_name}-{tetrad.nt3.full_name}-{tetrad.nt4.full_name}"
    )
    return [
        TetradPairDTO(
            id_(tp.tetrad1),
            id_(tp.tetrad2),
            tp.direction.value,
            float(tp.rise),
            float(tp.twist),
        )
        for tp in tetrad_pairs
    ]


def convert_quadruplexes(quadruplexes: List[Quadruplex]) -> List[QuadruplexDTO]:
    nts_ = lambda nts: [nt.full_name for nt in nts]
    return [
        QuadruplexDTO(
            convert_tetrads(q),
            q.onzm.value if q.onzm else None,
            (
                LoopClassificationDTO(
                    q.loop_class.value, q.loop_class.loop_progression()
                )
                if q.loop_class
                else None
            ),
            [g.value for g in q.gba_classes],
            [
                nts_(q.tracts[0].nucleotides),
                nts_(q.tracts[1].nucleotides),
                nts_(q.tracts[2].nucleotides),
                nts_(q.tracts[3].nucleotides),
            ],
            [nt.full_name for nt in q.bulges],
            [
                LoopDTO(
                    l.loop_type.value if l.loop_type is not None else None,
                    nts_(l.nucleotides),
                )
                for l in q.loops
            ],
        )
        for q in quadruplexes
    ]


def convert_helices(analysis: Analysis) -> List[HelixDTO]:
    helices = []
    for h in analysis.helices:
        # take only quadruplexes with at least two tetrads
        quadruplexes = [q for q in h.quadruplexes if len(q.tetrads) >= 2]
        # gather tetrads from selected quadruplexes
        tetrads = set([t for q in quadruplexes for t in q.tetrads])
        # gather pairs among all selected tetrads
        tetrad_pairs = [
            tp
            for tp in h.tetrad_pairs
            if tp.tetrad1 in tetrads and tp.tetrad2 in tetrads
        ]

        if quadruplexes and tetrad_pairs:
            helices.append(
                HelixDTO(
                    convert_quadruplexes(quadruplexes),
                    convert_tetrad_pairs(tetrad_pairs),
                )
            )
    return helices


def convert_dot_bracket(analysis: Analysis) -> TwoLineDotBracketDTO:
    return TwoLineDotBracketDTO(analysis.sequence, analysis.line1, analysis.line2)


def convert_quadruplex_dot_bracket(analysis: Analysis) -> QuadruplexDotBracketDTO:
    letters = "QRSTUVWXYZ" + "".join(reversed("ABCDEFGHIJKLMNOP"))

    if len(analysis.tetrads) > len(letters):
        raise ValueError(f"Cannot represent more than {len(letters)} tetrads")

    lines = analysis.mapping.dot_bracket.splitlines()
    sequences = []
    structures = []

    i = 0
    while i < len(lines):
        sequences.append(lines[i + 1])
        structures.append(lines[i + 2])
        i += 3

    mapping = {}
    i = 0
    current_chain = None

    for residue in analysis.structure3d.residues:
        if residue.is_nucleotide:
            if current_chain is None:
                current_chain = residue.chain
            if residue.chain != current_chain:
                current_chain = residue.chain
                i += 1
            mapping[residue] = i
            i += 1

    sequence = "-".join(sequences)
    structure = list("-".join(structures))
    assert len(sequence) == len(structure)
    assert len(sequence) == i

    for i, tetrad in enumerate(analysis.tetrads):
        nt1, nt2, nt3, nt4 = sorted(tetrad.nucleotides)
        if tetrad.onz == ONZ.O_PLUS:
            structure[mapping[nt1]] = letters[i]
            structure[mapping[nt2]] = letters[i].lower()
            structure[mapping[nt3]] = letters[i]
            structure[mapping[nt4]] = letters[i].lower()
        elif tetrad.onz == ONZ.O_MINUS:
            structure[mapping[nt1]] = letters[i].lower()
            structure[mapping[nt2]] = letters[i]
            structure[mapping[nt3]] = letters[i].lower()
            structure[mapping[nt4]] = letters[i]
        elif tetrad.onz == ONZ.N_PLUS:
            structure[mapping[nt1]] = letters[i]
            structure[mapping[nt2]] = letters[i].lower()
            structure[mapping[nt3]] = letters[i].lower()
            structure[mapping[nt4]] = letters[i]
        elif tetrad.onz == ONZ.N_MINUS:
            structure[mapping[nt1]] = letters[i].lower()
            structure[mapping[nt2]] = letters[i]
            structure[mapping[nt3]] = letters[i]
            structure[mapping[nt4]] = letters[i].lower()
        elif tetrad.onz == ONZ.Z_PLUS:
            structure[mapping[nt1]] = letters[i]
            structure[mapping[nt2]] = letters[i]
            structure[mapping[nt3]] = letters[i].lower()
            structure[mapping[nt4]] = letters[i].lower()
        elif tetrad.onz == ONZ.Z_MINUS:
            structure[mapping[nt1]] = letters[i].lower()
            structure[mapping[nt2]] = letters[i].lower()
            structure[mapping[nt3]] = letters[i]
            structure[mapping[nt4]] = letters[i]
        else:
            raise ValueError(f"Unexpected ONZ value: {tetrad.onz}")

    chi_line = list(re.sub(r"[^-]", ".", sequence))

    for nt in analysis.structure3d.residues:
        if nt.is_nucleotide and nt in analysis.global_index:
            chi_line[mapping[nt]] = (
                "a"
                if nt.chi_class == GlycosidicBond.anti
                else "s"
                if nt.chi_class == GlycosidicBond.syn
                else "?"
            )

    loop_line = list(re.sub(r"[^-]", ".", sequence))

    for helix in analysis.helices:
        for quadruplex in helix.quadruplexes:
            for loop in quadruplex.loops:
                for nt in loop.nucleotides:
                    loop_line[mapping[nt]] = (
                        "d"
                        if loop.loop_type == LoopType.diagonal
                        else "P"
                        if loop.loop_type == LoopType.propeller_plus
                        else "p"
                        if loop.loop_type == LoopType.propeller_minus
                        else "L"
                        if loop.loop_type == LoopType.lateral_plus
                        else "l"
                        if loop.loop_type == LoopType.lateral_minus
                        else "?"
                    )
            # mark bulges
            for nt in quadruplex.bulges:
                loop_line[mapping[nt]] = "b"

    return QuadruplexDotBracketDTO(
        sequence, "".join(structure), "".join(chi_line), "".join(loop_line)
    )


def generate_dto(analysis: Analysis):
    metals = convert_metals(analysis)
    nucleotides = convert_nucleotides(analysis)
    base_pairs = convert_base_pairs(analysis)
    helices = convert_helices(analysis)
    dot_bracket = convert_dot_bracket(analysis)
    quadruplex_dot_bracket = convert_quadruplex_dot_bracket(analysis)
    return ResultDTO(
        metals, nucleotides, base_pairs, helices, dot_bracket, quadruplex_dot_bracket
    )
